/*
 *  client.c -- Outbound calls
 *
 *  client.c is a part of binkd project
 *
 *  Copyright (C) 1996  Dima Maloff, 5047/13
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. See COPYING.
 */

/*
 * $IdS: client.c,v 1.1 1996/12/02 04:40:44 mff Exp mff $
 *
 * $Log: client.c,v $
 * Revision 1.4  1996/12/14  07:03:26  mff
 * Now we use branch()
 *
 * Revision 1.3  1996/12/09  03:37:21  mff
 * Changed call-out Logic
 *
 * Revision 1.2  1996/12/07  12:03:02  mff
 * Now fork()'s or _beginthread()'s
 *
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef HAVE_FORK
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#elif defined(HAVE_THREADS)
#include <dos.h>
#include <process.h>
#else
#error Must define either HAVE_FORK or HAVE_THREADS!
#endif

#include "sys.h"
#include "client.h"
#include "Config.h"
#include "readcfg.h"
#include "iphdr.h"
#include "ftnq.h"
#include "tools.h"
#include "protocol.h"
#include "bsy.h"

static void call (void *arg);

int n_clients = 0;

#ifdef HAVE_FORK

static void chld (int pid)
{
  --n_clients;
  pid = -1;
  wait (&pid);
}

#endif

void clientmgr (void *arg)
{
  FTNQ *q = 0;
  extern int pidcmgr;
  int pid;

#ifdef HAVE_FORK
  signal (SIGCHLD, chld);
#endif

  pidcmgr = 0;

  while (1)
  {
    FTN_NODE *r;

    if (!q)
    {
      q = q_scan ();
      q_list (stderr, q);
    }
    if (n_clients < max_clients)
    {
      if ((r = q_next_node (q)) != 0)
      {
	++n_clients;
	if ((pid = branch (call, (void *) DUP (*r))) < 0)
	{
	  --n_clients;
	  Log (0, "cannot branch out");
	}
	else
	{
	  Log (5, "started client #%i, id=%i", n_clients, pid);
	}
      }
      else
      {
	q_free (q);
	q = 0;
	sleep (rescan_delay);
      }
    }
    else
    {
      sleep (call_delay);
    }
  }
}

static int call0 (FTN_NODE *node)
{
  int sockfd;
  struct hostent he;
  struct hostent *hp;
  struct sockaddr_in sin;
  struct in_addr defaddr;
  char **cp;
  char szDestAddr[FTN_ADDR_SZ + 1];

  ftnaddress_to_str (szDestAddr, &node->fa);
  Log (2, "call to %s", szDestAddr);

  sin.sin_port = htons (oport);

  if (!isdigit (node->host[0]) ||
      (defaddr.s_addr = inet_addr (node->host)) == -1)
  {
    /* If not a raw ip address, try nameserver */
    if (!(hp = gethostbyname (node->host)))
    {
      Log (1, "%s: unknown host", node->host);
      return 0;
    }
  }
  else
  {
    /* Raw ip address, fake  */
    char *alist[2];

    hp = &he;
    hp->h_name = node->host;		    /* (char *) strdup (node->host); */
    hp->h_aliases = 0;
    hp->h_addrtype = AF_INET;
    hp->h_length = sizeof (struct in_addr);

    hp->h_addr_list = alist;		    /* (char **) malloc (2 * sizeof *
					     * (char *)); */
    hp->h_addr_list[0] = (char *) &defaddr;
    hp->h_addr_list[1] = (char *) 0;
  }
  sin.sin_family = hp->h_addrtype;

  /* Trying... */

  for (cp = hp->h_addr_list; cp && *cp; cp++)
  {
    if ((sockfd = socket (hp->h_addrtype, SOCK_STREAM, 0)) < 0)
    {
      Log (1, "socket: %s", TCPERR ());
      return 0;
    }
    sin.sin_addr = *((struct in_addr *) * cp);
    Log (4, "trying %s...", inet_ntoa (sin.sin_addr));
    if (!connect (sockfd, (struct sockaddr *) & sin, sizeof (sin)))
    {
      Log (4, "connected");
      break;
    }

    Log (1, "unable to connect: %s", TCPERR ());
    soclose (sockfd);
  }
  if (!cp || !*cp)
    return 0;

  protocol (sockfd, node);
  soclose (sockfd);
  return 1;
}

static void call (void *arg)
{
  FTN_NODE *node = (FTN_NODE *) arg;

  if (bsy_add (&node->fa))
  {
    call0 (node);
    bsy_remove (&node->fa);
  }
  free (arg);
#ifdef HAVE_THREADS
  --n_clients;
#endif
}
