#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(__WATCOMC__) || defined(VISUALCPP)
#include <direct.h>
#include <sys/utime.h>
#else
#include <dirent.h>
#include <utime.h>
#endif

#include "Config.h"
#include "sys.h"
#include "readcfg.h"
#include "inbound.h"
#include "tools.h"
#ifdef VISUALCPP
#include "win95/dirwin32.h"
#endif
#include "protocol.h"

#define INBOUND ((secure == P_SECURE) ? inbound : inbound_nonsecure)

/*
 * Searches for the ``file'' in the inbound and returns it's tmp name in s.
 * S must have MAXPATHLEN chars. Returns 0 on error, 1=found, 2=created.
 */
static int find_tmp_name (char *s, char *file, unsigned long size,
			   time_t time, FTN_ADDR *from, int secure)
{
  char buf[MAXPATHLEN + 80];
  char node[FTN_ADDR_SZ + 1];
  DIR *dp;
  struct dirent *de;
  FILE *f;
  int i, found = 0;
  char *t;

  if ((dp = opendir (INBOUND)) == 0)
  {
    Log (1, "cannot opendir %s: %s", INBOUND, strerror (errno));
    return 0;
  }
  strnzcpy (s, INBOUND, MAXPATHLEN);
  strnzcat (s, PATH_SEPARATOR, MAXPATHLEN);
  t = s + strlen (s);
  while ((de = readdir (dp)) != 0)
  {
    for (i = 0; i < 8; ++i)
      if (!isxdigit (de->d_name[i]))
	break;
    if (i < 8 || stricmp (de->d_name + 8, ".hr"))
      continue;
    strnzcat (s, de->d_name, MAXPATHLEN);
    if ((f = fopen (s, "r")) == 0 || !fgets (buf, sizeof (buf), f))
    {
      Log (1, "find_tmp_name: %s: %s", de->d_name, strerror (errno));
      if (f)
	fclose (f);
    }
    else
    {
      char *w[4];
      FTN_ADDR fa;

      FA_ZERO (&fa);
      for (i = 0; i < 4; ++i)
	w[i] = getwordx (buf, i + 1, 0);

      if (!strcmp (w[0], file) &&
	  size == atol (w[1]) &&
	  (time & ~1) == (atol (w[2]) & ~1) &&
	  parse_ftnaddress (w[3], &fa) && !ftnaddress_cmp (&fa, from))
      {
	found = 1;
      }
      for (i = 0; i < 4; ++i)
	if (w[i])
	  free (w[i]);
      fclose (f);
    }
    if (found)
      break;
    else
      *t = 0;
  }
  closedir (dp);

  /* New file */
  if (!found)
  {
    char tmp[20];

    Log (5, "File not found, trying to create a tmpname");
    while (1)
    {
      sprintf (tmp, "%08lx.hr", rnd ());
      strnzcat (s, tmp, MAXPATHLEN);
      if (create_sem_file (s))
      {
	if ((f = fopen (s, "w")) == 0)
	{
	  Log (1, "%s: %s", s, strerror (errno));
	  return 0;
	}
	ftnaddress_to_str (node, from);
	fprintf (f, "%s %li %li %s\n", file, size, time, node);
	fclose (f);
	found = 2;
	break;
      }
    }
  }

  /* Replacing .hr with .dt */
  strcpy (strrchr (s, '.'), ".dt");
  return found;
}

FILE *inb_fopen (char *netname, unsigned long size,
		  time_t time, FTN_ADDR *from, int secure)
{
  char buf[MAXPATHLEN + 1];
  FILE *f;

  if (!find_tmp_name (buf, netname, size, time, from, secure))
    return 0;

  if ((f = fopen (buf, "ab")) == 0)
    Log (1, "%s: %s", buf, strerror (errno));
  return f;
}

int inb_done (char *netname, unsigned long size,
	       time_t time, FTN_ADDR *from, int secure)
{
  char tmp_name[MAXPATHLEN + 1];
  char real_name[MAXPATHLEN + 1];
  char *s;
  struct utimbuf utb;

  if (find_tmp_name (tmp_name, netname, size, time, from, secure) != 1)
    return 0;

  strnzcpy (real_name, INBOUND, MAXPATHLEN);
  strnzcat (real_name, PATH_SEPARATOR, MAXPATHLEN);
  s = real_name + strlen (real_name);
  strnzcat (real_name, netname, MAXPATHLEN);
  strwipe (s);
  s = real_name + strlen (real_name) - 1;
  while (1)
  {
    if (!RENAME (tmp_name, real_name))
    {
      Log (2, "%s -> %s", netname, real_name);
      break;
    }
    else
      Log (5, "error renaming `%s' to `%s': %s",
	   netname, real_name, strerror (errno));
    if (isalpha (*s) && toupper (*s) != 'Z')
      ++* s;
    else if (isdigit (*s) && toupper (*s) != '9')
      ++* s;
    else if (*--s == '.' || *s == '\\' || *s == '/')
    {
      Log (1, "cannot rename %s to it's realname! (data stored in %s)",
	   netname, tmp_name);
      return 0;
    }
  }

  utb.actime = utb.modtime = time;
  if (utime (real_name, &utb) == -1)
    Log (1, "utime: %s", strerror (errno));

  /* Replacing .dt with .hr and removing temp. file */
  strcpy (strrchr (tmp_name, '.'), ".hr");
  if (unlink (tmp_name) == -1)
    Log (1, "error unlinking `%s': %s", tmp_name, strerror (errno));

  return 1;
}

/*
 * Checks if the file is already exists in our inbound
 */
int inb_test (char *filename, unsigned long size, time_t t, int secure)
{
  char fp[MAXPATHLEN + 1];
  char *s;
  struct stat sb;

  strnzcpy (fp, INBOUND, MAXPATHLEN);
  strnzcat (fp, PATH_SEPARATOR, MAXPATHLEN);
  s = fp + strlen (fp);
  strnzcat (fp, filename, MAXPATHLEN);
  strwipe (s);

  return stat (fp, &sb) == 0 && sb.st_size == size &&
    (sb.st_mtime & ~1) == (t & ~1);
}
